// -*- mode: c++; c-basic-offset: 2; indent-tabs-mode: nil; -*-
//
// Quick hack based on ffmpeg
// tutorial http://dranger.com/ffmpeg/tutorial01.html
// in turn based on a tutorial by
// Martin Bohme (boehme@inb.uni-luebeckREMOVETHIS.de)
//

// Ancient AV versions forgot to set this.
#define __STDC_CONSTANT_MACROS

// libav: "U NO extern C in header ?"
extern "C"
{
#include <libavcodec/avcodec.h>
#include <libavformat/avformat.h>
#include <libswscale/swscale.h>
}

#include "content-streamer.h"
#include "led-matrix.h"
#include <fcntl.h>
#include <getopt.h>
#include <limits.h>
#include <signal.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

using rgb_matrix::FrameCanvas;
using rgb_matrix::RGBMatrix;
using rgb_matrix::StreamIO;
using rgb_matrix::StreamWriter;

volatile bool interrupt_received = false;
static void InterruptHandler(int) {
	interrupt_received = true;
}

struct LedPixel {
	uint8_t r, g, b;
};

void CopyFrame(AVFrame *pFrame, FrameCanvas *canvas) {
	// Write pixel data
	const int height = canvas->height();
	const int width	 = canvas->width();
	for (int y = 0; y < height; ++y) {
		LedPixel *pix = (LedPixel *) (pFrame->data[0] + y * pFrame->linesize[0]);
		for (int x = 0; x < width; ++x, ++pix) { canvas->SetPixel(x, y, pix->r, pix->g, pix->b); }
	}
}

static void add_nanos(struct timespec *accumulator, long nanoseconds) {
	accumulator->tv_nsec += nanoseconds;
	while (accumulator->tv_nsec > 1000000000) {
		accumulator->tv_nsec -= 1000000000;
		accumulator->tv_sec += 1;
	}
}

int main(int argc, char *argv[]) {
	RGBMatrix::Options matrix_options;
	rgb_matrix::RuntimeOptions runtime_opt;

	// Initalizing these to NULL prevents segfaults!
	AVFormatContext *pFormatCtx	  = NULL;
	AVCodecContext *pCodecCtxOrig = NULL;
	AVCodecContext *pCodecCtx	  = NULL;
	AVCodec *pCodec				  = NULL;
	AVFrame *pFrame				  = NULL;
	AVFrame *pFrameRGB			  = NULL;
	uint8_t *buffer				  = NULL;
	struct SwsContext *sws_ctx	  = NULL;
	AVPacket packet;
	int i;
	int videoStream;
	int frameFinished;
	int numBytes;

	const char *movie_file = argv[optind];

	// Register all formats and codecs
	av_register_all();
	avformat_network_init();

	// Open video file
	if (avformat_open_input(&pFormatCtx, movie_file, NULL, NULL) != 0) return -1; // Couldn't open file

	// Retrieve stream information
	if (avformat_find_stream_info(pFormatCtx, NULL) < 0) return -1; // Couldn't find stream information

	long frame_count  = 0;
	RGBMatrix *matrix = CreateMatrixFromOptions(matrix_options, runtime_opt);
	if (matrix == NULL) { return 1; }

	FrameCanvas *offscreen_canvas = matrix->CreateFrameCanvas();
	// Find the first video stream
	videoStream = -1;
	for (i = 0; i < (int) pFormatCtx->nb_streams; ++i) {
		if (pFormatCtx->streams[i]->codec->codec_type == AVMEDIA_TYPE_VIDEO) {
			videoStream = i;
			break;
		}
	}
	if (videoStream == -1) return -1; // Didn't find a video stream

	// Get a pointer to the codec context for the video stream
	pCodecCtxOrig = pFormatCtx->streams[videoStream]->codec;
	double fps	  = av_q2d(pFormatCtx->streams[videoStream]->avg_frame_rate);
	if (fps < 0) { fps = 1.0 / av_q2d(pFormatCtx->streams[videoStream]->codec->time_base); }

	// Find the decoder for the video stream
	pCodec = avcodec_find_decoder(pCodecCtxOrig->codec_id);
	if (pCodec == NULL) {
		fprintf(stderr, "Unsupported codec!\n");
		return -1;
	}
	// Copy context
	pCodecCtx = avcodec_alloc_context3(pCodec);
	if (avcodec_copy_context(pCodecCtx, pCodecCtxOrig) != 0) {
		fprintf(stderr, "Couldn't copy codec context");
		return -1;
	}

	// Open codec
	if (avcodec_open2(pCodecCtx, pCodec, NULL) < 0) return -1;

	// Allocate video frame
	pFrame = av_frame_alloc();

	// Allocate an AVFrame structure
	pFrameRGB = av_frame_alloc();
	if (pFrameRGB == NULL) return -1;

	// Determine required buffer size and allocate buffer
	numBytes = avpicture_get_size(AV_PIX_FMT_RGB24, pCodecCtx->width, pCodecCtx->height);
	buffer	 = (uint8_t *) av_malloc(numBytes * sizeof(uint8_t));

	// Assign appropriate parts of buffer to image planes in pFrameRGB
	// Note that pFrameRGB is an AVFrame, but AVFrame is a superset
	// of AVPicture
	avpicture_fill((AVPicture *) pFrameRGB, buffer, AV_PIX_FMT_RGB24, pCodecCtx->width, pCodecCtx->height);

	// initialize SWS context for software scaling
	sws_ctx = sws_getContext(
	  pCodecCtx->width,
	  pCodecCtx->height,
	  pCodecCtx->pix_fmt,
	  matrix->width(),
	  matrix->height(),
	  AV_PIX_FMT_RGB24,
	  SWS_BILINEAR,
	  NULL,
	  NULL,
	  NULL);
	if (sws_ctx == 0) {
		fprintf(stderr, "Trouble doing scaling to %dx%d :(\n", matrix->width(), matrix->height());
		return 1;
	}

	signal(SIGTERM, InterruptHandler);
	signal(SIGINT, InterruptHandler);

	const long frame_wait_nanos = 1e9 / fps;
	struct timespec next_frame;

	do {
		unsigned int frames_left	= UINT_MAX;
		unsigned int frames_to_skip = 0;
		av_seek_frame(pFormatCtx, videoStream, 0, AVSEEK_FLAG_ANY);
		avcodec_flush_buffers(pCodecCtx);
		clock_gettime(CLOCK_MONOTONIC, &next_frame);
		while (!interrupt_received && av_read_frame(pFormatCtx, &packet) >= 0 && frames_left > 0) {
			// Is this a packet from the video stream?
			if (packet.stream_index == videoStream) {
				if (frames_to_skip) {
					frames_to_skip--;
					continue;
				}

				// Determine absolute end of this frame now so that we don't include
				// decoding overhead. TODO: skip frames if getting too slow ?
				add_nanos(&next_frame, frame_wait_nanos);

				// Decode video frame
				avcodec_decode_video2(pCodecCtx, pFrame, &frameFinished, &packet);

				// Did we get a video frame?
				if (frameFinished) {
					// Convert the image from its native format to RGB
					sws_scale(
					  sws_ctx,
					  (uint8_t const *const *) pFrame->data,
					  pFrame->linesize,
					  0,
					  pCodecCtx->height,
					  pFrameRGB->data,
					  pFrameRGB->linesize);

					CopyFrame(pFrameRGB, offscreen_canvas);
					frame_count++;
					frames_left--;
					offscreen_canvas = matrix->SwapOnVSync(offscreen_canvas);
				}
			}
			// Free the packet that was allocated by av_read_frame
			av_free_packet(&packet);
		}
	} while (!interrupt_received);

	if (interrupt_received) {
		// Feedback for Ctrl-C, but most importantly, force a newline
		// at the output, so that commandline-shell editing is not messed up.
		fprintf(stderr, "Got interrupt. Exiting\n");
	}

	delete matrix;

	// Free the RGB image
	av_free(buffer);
	av_frame_free(&pFrameRGB);

	// Free the YUV frame
	av_frame_free(&pFrame);

	// Close the codecs
	avcodec_close(pCodecCtx);
	avcodec_close(pCodecCtxOrig);

	// Close the video file
	avformat_close_input(&pFormatCtx);

	fprintf(stderr, "Total of %ld frames decoded\n", frame_count);

	return 0;
}
